#ifndef __RAW2NEF_H
#define __RAW2NEF_H
//
#define CORE_VERSION "v0.15"

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;
typedef unsigned short color_t;

struct unpack_buff { char buf[12]; };
const char* unpack( unpack_buff& buf, unsigned x );
unsigned pack( const char* s );

#define CAMERA_ID(id) pack(#id)

struct TCCDParam
{
   unsigned raw_width;
   unsigned raw_height;
   unsigned file_size;
   unsigned cfa_colors;
   unsigned row_length;
   unsigned bits_per_sample;
   unsigned camera;
   unsigned data_offset;

   enum
   {
      GMYC, MGCY, YCGM, CYMG,
      BGGR, GBRG, GRBG, RGGB,
      ShiftRightBit = 1,
      ShiftDownBit = 2,
      PrimaryColorsBit = 4,
   };
   enum
   {
      fInterlaced       = 0x00000001,
      fARM              = 0x00000002,
      fIntelByteOrder   = 0x00000004,
   };

   unsigned flags;
};

enum EReadResult
{
   e_success,
   e_unexpected_eof, // incomplete file ?
   e_too_large_file, // bad format description ?
   e_not_supported, // format not supported
   e_error_format, // unexpected format
   e_exception, // unexpected error in reading
};

int read_raw(FILE* in, color_t ccd[], TCCDParam& ccd_pars);


#ifndef __TIFF_H
#include "tiff.h"
#endif

template <class T>
class buff_t
{
public:
   buff_t() : buf(0) {}
   buff_t(unsigned size) : buf( new T[size]) {}
   ~buff_t() { delete [] buf; }

   void alloc(unsigned size) { delete [] buf; buf = new T[size]; }
   
   operator T* () const { return buf; }
   T* buf;
};

class TIFF_Content
{
public:
   TIFF_Content() 
   {
      ccd_pars.camera = 0;
      ccd_pars.flags = 0;
      ccd_pars.raw_width = 0;
      ccd_pars.raw_height = 0;
   }

   IFDir IFD0;
   IFDir EXIF;
   IFDir MakerNote;
   IFDir SubIFD[4];

   TCCDParam ccd_pars;
   buff_t<color_t> picture;

   bool Is_NIKON();

   IFDir* Is_CFA();
   bool read_exif(FILE*);
   bool read_tiff(FILE*);
   int read_raw(FILE*, IFDir*);

   bool write_nef(FILE*, bool, bool=false);
   bool write_dng(FILE*, bool, bool=true, bool=false);

   bool mask_dead_pixels(FILE*);

private:
   bool safe_read_exif(FILE*);
   bool safe_read_tiff(FILE*);
   int safe_read_raw(FILE*, IFDir*);
};

struct TRaw2Nef
{
   TRaw2Nef();
   ~TRaw2Nef();

   char* ini_fn; // ini-file name
   char* ifn; // input RAW file name
   char* ofn; // output NEF file name
   char* efn; // EXIF file name
   char* dfn; // dead pixels map file name

   char* exe_fn; // this program path and filename

   char* ofn_buf;
   char* efn_buf;
   char* ini_buf;
   char* dfn_buf;

   bool verbose;
   bool copy_exif; // copy EXIF metadata from paired JPEG to NEF
   bool endian;   // endian: true for Intel, false for Motorola
   bool compatible;
   bool optimize; // try to minimize DNG file size (for comparing)
   bool incomplete; // allows load incomplete raw files
   bool same_number; // EXIF file has the same number as RAW file

   bool black_pixels;
   bool bright_pixels;
   bool append_table;
   int threashold;

   bool ProcessFile(unsigned camera);
   bool ParseIniFile(TCCDParam& ccd_pars, unsigned camera, unsigned file_size);
   void CreateExifFileName(const char* FilePath, int ExtOffset);

   bool raw_to_nef(int n, char* a[]);
   bool process_data(TIFF_Content& );
};

#endif
