#include <stdio.h>
#include <string.h>
#include <excpt.h>
#pragma hdrstop

/*
   Coolpix RAW to Coolpix NEF/DNG converter
   by paul69, v0.01, 2006-01-09, <e2500@narod.ru>

   Reading CFA raw file
*/

#include "tiff.h"
#include "raw2nef.h"

bool read_jpeg( FILE* in, color_t Pix[], bool bug );

int read_cfa( FILE* in, IFDir* IFD, TIFF_Content& NEF )
{
   IFDir::Tag* PhotometricInterpretation = IFD->get_tag(TIFF::PhotometricInterpretation);
   if( !PhotometricInterpretation ) return e_error_format;
   if( PhotometricInterpretation->get_value() != 0x8023 ) return e_error_format;

   IFDir::Tag* CFARepeatPatternDim = IFD->get_tag(TIFF::CFARepeatPatternDim);
   if( !CFARepeatPatternDim ) return e_error_format;
   if( CFARepeatPatternDim->count != 2 ) return e_error_format;
   if( CFARepeatPatternDim->get_value(0) != 2 ) return e_error_format;
   if( CFARepeatPatternDim->get_value(1) != 2 ) return e_error_format;

   IFDir::Tag* CFAPattern = IFD->get_tag(TIFF::CFAPattern);
   if( !CFAPattern ) return e_error_format;
   if( CFAPattern->count != 4 ) return e_error_format;
   unsigned cfa = (CFAPattern->get_value(0) << 16)
      | (CFAPattern->get_value(1) << 12)
      | (CFAPattern->get_value(2) << 8)
      | (CFAPattern->get_value(3) << 4);

   TCCDParam& ccd_pars = NEF.ccd_pars;

   switch(cfa)
   {
   case 0x14530: ccd_pars.cfa_colors = TCCDParam::GMYC; break;
   case 0x41350: ccd_pars.cfa_colors = TCCDParam::MGCY; break;
   case 0x53140: ccd_pars.cfa_colors = TCCDParam::YCGM; break;
   case 0x35410: ccd_pars.cfa_colors = TCCDParam::CYMG; break;
   case 0x21100: ccd_pars.cfa_colors = TCCDParam::BGGR; break;
   case 0x12010: ccd_pars.cfa_colors = TCCDParam::GBRG; break;
   case 0x10210: ccd_pars.cfa_colors = TCCDParam::GRBG; break;
   case 0x01120: ccd_pars.cfa_colors = TCCDParam::RGGB; break;
   default: ccd_pars.cfa_colors = cfa;
      if( ((cfa >> 16) & 0x0F) <= 2
       && ((cfa >> 12) & 0x0F) <= 2
       && ((cfa >>  8) & 0x0F) <= 2
       && ((cfa >>  4) & 0x0F) <= 2 )
         ccd_pars.cfa_colors |= TCCDParam::PrimaryColorsBit;
      break;
   }

   unsigned Width = IFD->get_tag(TIFF::ImageWidth)->get_value();
   unsigned Height = IFD->get_tag(TIFF::ImageHeight)->get_value();
   
   ccd_pars.raw_width = Width;
   ccd_pars.raw_height = Height;
   ccd_pars.file_size = 0;
   ccd_pars.row_length = 0;
   ccd_pars.bits_per_sample = 0;
   ccd_pars.camera = 0;
   ccd_pars.data_offset = 0;
   ccd_pars.flags = 0;

   IFDir::Tag* Compression = IFD->get_tag(TIFF::Compression);
   if( !Compression ) return e_error_format;
   if( Compression->get_value() == 7 )
   {
      // JPEG
      bool bug = false;
      IFDir::Tag* DNGVersion = NEF.IFD0.get_tag(TIFF::DNGVersion);
      if( DNGVersion )
      {
         unsigned version = (DNGVersion->get_value(0) << 24)
          | (DNGVersion->get_value(1) << 16)
          | (DNGVersion->get_value(2) << 8)
          | (DNGVersion->get_value(3));

         if( version == 0x01000000 )
            bug = true;
      }
      
      IFDir::Tag* BitsPerSample = IFD->get_tag(TIFF::BitsPerSample);
      if( !BitsPerSample ) return e_error_format;
      if( BitsPerSample->count != 1 ) return e_error_format;
      ccd_pars.bits_per_sample = BitsPerSample->get_value();

      IFDir::Tag* StripOffsets = IFD->get_tag(TIFF::StripOffsets);
      if( StripOffsets )
      {
         IFDir::Tag* StripByteCounts = IFD->get_tag(TIFF::StripByteCounts);
         if( !StripByteCounts ) return e_error_format;
         if( StripOffsets->count != StripByteCounts->count ) return e_error_format;

         IFDir::Tag* RowsPerStrip = IFD->get_tag(TIFF::RowsPerStrip);
         if( !RowsPerStrip ) return e_error_format;


         return e_not_supported;
      }

      IFDir::Tag* TileOffsets = IFD->get_tag(TIFF::TileOffsets);
      if( !TileOffsets ) return e_error_format;

      IFDir::Tag* TileByteCounts = IFD->get_tag(TIFF::TileByteCounts);
      if( !TileByteCounts ) return e_error_format;
      if( TileOffsets->count != TileByteCounts->count ) return e_error_format;

      IFDir::Tag* TileWidth = IFD->get_tag(TIFF::TileWidth);
      if( !TileWidth ) return e_error_format;

      IFDir::Tag* TileLength = IFD->get_tag(TIFF::TileLength);
      if( !TileLength ) return e_error_format;

      unsigned tile_w = TileWidth->get_value();
      unsigned tile_h = TileLength->get_value();
      unsigned tile_size = tile_w * tile_h;
      if( !tile_size ) return e_error_format;

      NEF.picture.alloc(Width*Height);

      buff_t<color_t> tile( tile_size );
      
      int tiles = TileOffsets->count;
      int tile_x = 0;
      int tile_y = 0;
      for(int tile_i=0; tile_i<tiles; ++tile_i)
      {
         fseek(in, TileOffsets->get_value(tile_i), SEEK_SET); 

         if( !read_jpeg( in, tile, bug ) )
            return e_unexpected_eof;

         int h = tile_h;
         if( tile_y + tile_h > Height )
            h = Height - tile_y;

         color_t* p = &NEF.picture[tile_y*Width + tile_x];

         int w = tile_w;
         if( tile_x + tile_w > Width )
            w = Width - tile_x;

         color_t* t = tile;

         for(int i=0; i<h; ++i, p += Width, t += tile_w )
         {
            for(int j=0; j<w; ++j )
            {
               p[j] = t[j];
            }
         }

         tile_x += tile_w;
         if( tile_x >= Width )
         {
            tile_x = 0;
            tile_y += tile_h;
         }
      }
      return e_success;
   }
   else
   if( Compression->get_value() == 1 )
   {
      // Uncompressed
      IFDir::Tag* BitsPerSample = IFD->get_tag(TIFF::BitsPerSample);
      if( !BitsPerSample ) return e_error_format;
      if( BitsPerSample->count != 1 ) return e_error_format;

      ccd_pars.bits_per_sample = BitsPerSample->get_value();
      if( ccd_pars.bits_per_sample == 12 )
      {
         IFDir::Tag* StripOffsets = IFD->get_tag(TIFF::StripOffsets);
         if( !StripOffsets )
            return e_error_format;

         IFDir::Tag* StripByteCounts = IFD->get_tag(TIFF::StripByteCounts);
         if( !StripByteCounts ) return e_error_format;
         if( StripOffsets->count != StripByteCounts->count ) return e_error_format;

         IFDir::Tag* RowsPerStrip = IFD->get_tag(TIFF::RowsPerStrip);
         if( !RowsPerStrip ) return e_error_format;

         IFDir::Tag* Make = NEF.IFD0.get_tag(TIFF::Make);
         if( Make && Make->count >= 5
          && 0 == memcmp( Make->value, "NIKON", 5 ) )
         {
            // restore original Make and Model
            if( Make->value[5] == 0 && Make->count > 7 )
            {
               int len = strlen(Make->value + 6);
               if( len > 0 && Make->count > len+6 )
               {
                  NEF.IFD0.add_ASCII(TIFF::Make, Make->value + 6, len );

                  IFDir::Tag* Model = NEF.IFD0.get_tag(TIFF::Model);
                  if( Model )
                  {
                     len = strlen(Model->value);
                     if( len == 5 && Model->count > 7 )
                     {
                        len = strlen(Model->value + 6);
                        if( len > 0 && Model->count > len+6 )
                           NEF.IFD0.add_ASCII(TIFF::Model, Model->value + 6, len );
                     }
                  }
               }
            }

            unsigned picture_size = Width*Height;
            NEF.picture.alloc( picture_size );

            ccd_pars.flags |= TCCDParam::fInterlaced;
            ccd_pars.cfa_colors &= TCCDParam::PrimaryColorsBit; // set GMYC or BGGR
            ccd_pars.row_length = (Width * 12 + 8-1) / 8;;
            ccd_pars.data_offset = StripOffsets->get_value();
            ccd_pars.file_size = ccd_pars.data_offset + StripByteCounts->get_value();
            ccd_pars.camera = 1;

            return read_raw(in, NEF.picture, ccd_pars);
         }
      }
      return e_error_format;
   }

   return e_not_supported;
}

int TIFF_Content::safe_read_raw(FILE* in, IFDir* CFA)
{
   try
   {
      return read_cfa( in, CFA, *this );
   }
   catch(const char* x)
   {
      printf("error: %s\n", x);
   }
   catch(...)
   {
      printf("error: exception\n");
   }

   return e_exception;
}

int TIFF_Content::read_raw(FILE* in, IFDir* CFA)
{
   try
   {
      return safe_read_raw( in, CFA );
   }
   __except(1)
   {
      printf("error: except\n");
   }
   return e_exception;
}


