#include <stdio.h>
#include <string.h>
#pragma hdrstop

/*
   Coolpix RAW to Coolpix NEF/DNG converter
   by paul69, v0.14, 2005-12-20, <e2500@narod.ru>

   Writting TIFF file
*/

#include <math.h> // sqrt
#include "tiff.h"
#include "raw2nef.h"

bool write(IFDir& ifd, FILE* out, bool endian, unsigned tiff_start, unsigned next_ifd)
{
   unsigned size = ifd.get_size();
   char* buffer = new char[size];

   bool ret = false;

   unsigned ifd_start = ftell(out) - tiff_start;

   if( ifd.write_to_buffer( buffer, size, endian, ifd_start, next_ifd) )
   {
      ret = fwrite( buffer, size, 1, out) == 1;
   }
   delete [] buffer;
   return ret;
}

bool write_word(FILE* out, bool endian, unsigned w)
{
   if( endian )
   {
      if( fputc( (int)(w & 0xFF), out) == EOF ) return false;
      return fputc( (int)((w >> 8)&0xFF), out) != EOF;
   }

   if( fputc( (int)((w >> 8)&0xFF), out) == EOF ) return false;
   return fputc( (int)(w & 0xFF), out) != EOF;
}

bool write_dword(FILE* out, bool endian, unsigned dw)
{
   if( endian )
   {
      if( fputc( (int)((dw) & 0xFF), out) == EOF ) return false;
      if( fputc( (int)((dw >> 8) & 0xFF), out) == EOF ) return false;
      if( fputc( (int)((dw >> 16) & 0xFF), out) == EOF ) return false;
      return fputc( (int)((dw >> 24) & 0xFF), out) != EOF;
   }

   if( fputc( (int)((dw >> 24) & 0xFF), out) == EOF ) return false;
   if( fputc( (int)((dw >> 16) & 0xFF), out) == EOF ) return false;
   if( fputc( (int)((dw >> 8) & 0xFF), out) == EOF ) return false;
   return fputc( (int)((dw) & 0xFF), out) != EOF;
}


static char cfa_pattern[8][5] = {
   "GMYC", "MGCY", "YCGM", "CYMG",
   "BGGR", "GBRG", "GRBG", "RGGB",
};

bool ccd_interpolation(color_t ccd[], color_t image[], int w, int h, TCCDParam& ccd_pars)
{
   enum { Gr, Mg, Cy, Ye };
   enum { R, G, B };


   int cfa[4]; // index of color filter

   if( ccd_pars.cfa_colors & TCCDParam::PrimaryColorsBit )
      for(int i=0; i<4; ++i)
      {
         switch( cfa_pattern[ccd_pars.cfa_colors][i] )
         {
         case 'R': cfa[i] = R; break;
         case 'G': cfa[i] = G; break;
         case 'B': cfa[i] = B; break;
         default: return false;
         }
      }
   else
      for(int i=0; i<4; ++i)
      {
         switch( cfa_pattern[ccd_pars.cfa_colors][i] )
         {
         case 'G': cfa[i] = Gr; break;
         case 'M': cfa[i] = Mg; break;
         case 'C': cfa[i] = Cy; break;
         case 'Y': cfa[i] = Ye; break;
         default: return false;
         }
      }

   int height = ccd_pars.raw_height;
   int width = ccd_pars.raw_width;

   int dx = width/w;
   int dy = height/h;

   color_t* out = image;

   for(int y=0; y<h; ++y)
   {
      int _y = y * dy;
      for(int x=0; x<w; ++x, out += 3)
      {
         int c[4];

         int _x = x * dx;
         color_t* in = &ccd[_y*width + _x];
         // index in CFA
         int i = (_x&1) | ((_y&1)<<1);

         c[ cfa[i] ] = in[0];
         c[ cfa[i^1] ] = in[1];
         c[ cfa[i^2] ] = in[width];
         c[ cfa[i^3] ] = in[width+1];

         int RR, GG, BB;
         if( ccd_pars.cfa_colors & TCCDParam::PrimaryColorsBit )
         {
            RR = (int)c[R];
               if( RR < 0 ) RR=0; if( RR > 65535 ) RR = 65535;

            GG = (int)c[G];
               if( GG < 0 ) GG=0; if( GG > 65535 ) GG = 65535;

            BB = (int)c[B];
               if( BB < 0 ) BB=0; if( BB > 65535 ) BB = 65535;
         }
         else
         {
            float Y = (float)((c[Gr] + c[Mg] + c[Ye] + c[Cy])/4);
            float Cr = (float)((c[Mg] + c[Ye]) - (c[Gr] + c[Cy]));
            float Cb = (float)((c[Mg] + c[Cy]) - (c[Gr] + c[Ye]));

            RR = (int)(Y + 1.402f*Cr);
               if( RR < 0 ) RR=0; if( RR > 65535 ) RR = 65535;

            GG = (int)(Y - 0.34414f*Cb - 0.71414f*Cr);
               if( GG < 0 ) GG=0; if( GG > 65535 ) GG = 65535;

            BB = (int)(Y + 1.772f*Cb);
               if( BB < 0 ) BB=0; if( BB > 65535 ) BB = 65535;
         }

         out[ R ] = (color_t)RR;
         out[ G ] = (color_t)GG;
         out[ B ] = (color_t)BB;
      }
   }
   return true;
}

bool write_thumbnail(FILE* out, color_t ccd[], int w, int h, TCCDParam& ccd_pars)
{
   int size = w*h;
   buff_t<color_t> tn(size*3);

   ccd_interpolation( ccd, tn, w, h, ccd_pars);

   int max_color[3];
   max_color[0]=max_color[1]=max_color[2]=0;
   int min_color[3];
   min_color[0]=min_color[1]=min_color[2]=65536;

   for(int i=0; i<size; ++i)
   {
      for(int j=0; j<3; ++j)
      {
         int pixel = tn[i*3+j];
         if( pixel > max_color[j] )
            max_color[j] = pixel;
         if( pixel < min_color[j] )
            min_color[j] = pixel;
      }
   }

   if( max_color[0] == 0 ) max_color[0] = 65535;
   if( max_color[1] == 0 ) max_color[1] = 65535;
   if( max_color[2] == 0 ) max_color[2] = 65535;

   if( min_color[0] == max_color[0] ) min_color[0] = 0;
   if( min_color[1] == max_color[1] ) min_color[1] = 0;
   if( min_color[2] == max_color[2] ) min_color[2] = 0;

   for(int i=0; i<size; ++i)
   {
      for(int j=0; j<3; ++j)
      {
         int pixel = tn[i*3+j];

         if( max_color[j] > 0 )
         {
            double temp = (double)(pixel - min_color[j])/(double)(max_color[j] - min_color[j]);
            temp = temp > 0 ? sqrt(temp) : 0.;
            pixel = (int)(temp*300);
         }

         if( pixel > 255 ) pixel = 255;
         fputc(pixel, out);
      }
   }
   return true;
}

////////////////////////////////////////////////////////////////////////////////
// Writting IFD to buffer

class TOut
{
public:
   TOut( char* buf, int size, bool endian ) : buf_(buf), size_(size), off_(0), endian_(endian) {}

   char* buf_;
   int size_;
   int off_;
   bool endian_;

public:
   void put(int c)
   {
      if( off_ < size_ )
      {
         buf_[off_++] = (char)c;
      }
   }

   bool is_place() { return off_ < size_; }
   bool is_place(int n) { return off_+n <= size_; }

   bool write_byte(unsigned v);
   bool write_word(unsigned w);
   bool write_dword(unsigned dw);
};

bool TOut::write_byte(unsigned v)
{
   if( !is_place(1) ) return false;

   put((int)v);

   return  true;
}

bool TOut::write_word(unsigned w)
{
   if( !is_place(2) ) return false;

   if( endian_ )
   {
      put( (int)(w & 0xFF) );
      put( (int)((w >> 8)&0xFF) );
      return true;
   }

   put( (int)((w >> 8)&0xFF) );
   put( (int)(w & 0xFF) );
   return true;
}

bool TOut::write_dword(unsigned dw)
{
   if( !is_place(4) ) return false;

   if( endian_ )
   {
      put( (int)((dw) & 0xFF) );
      put( (int)((dw >> 8) & 0xFF) );
      put( (int)((dw >> 16) & 0xFF) );
      put( (int)((dw >> 24) & 0xFF) );
      return true;
   }

   put( (int)((dw >> 24) & 0xFF) );
   put( (int)((dw >> 16) & 0xFF) );
   put( (int)((dw >> 8) & 0xFF) );
   put( (int)((dw) & 0xFF) );
   return true;
}

bool IFDir::write_to_buffer( char* buf, int size, bool endian, unsigned ifd_start, unsigned next_ifd)
{
   TOut out( buf, size, endian );
   if( !out.write_word(tag_count_) )
      return false;

   int i, k;
   unsigned data_offset = ifd_start + 2 + 12*tag_count_ + 4;

   for(i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];

      out.write_word(t->tag);
      out.write_word(t->type);
      out.write_dword(t->count);

      unsigned tag_data_size = t->get_size();
      if( tag_data_size > 4 )
      {
         out.write_dword(data_offset);
         data_offset += tag_data_size;
      }
      else
      {
         switch(t->type)
         {
         case TIFF::type_BYTE:
         case TIFF::type_SBYTE:
         case TIFF::type_ASCII:
         case TIFF::type_UNDEFINED:
            for(k=0; k<4; ++k)
            {
               out.write_byte(k < t->count ? ((byte*)t->value)[k] : (byte)0 );
            }
            break;

         case TIFF::type_WORD:
         case TIFF::type_SHORT:
            for(k=0; k<2; ++k)
            {
               out.write_word(k < t->count ? ((word*)t->value)[k] : (word)0 );
            }
            break;

         case TIFF::type_DWORD:
         case TIFF::type_LONG:
         case TIFF::type_FLOAT:
            out.write_dword(*(dword*)t->value);
            break;
         }
      }
   }

   out.write_dword(next_ifd);

   for(i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];

      unsigned tag_data_size = t->get_size();
      if( tag_data_size > 4 )
      {
         switch(t->type)
         {
         case TIFF::type_BYTE:
         case TIFF::type_SBYTE:
         case TIFF::type_ASCII:
         case TIFF::type_UNDEFINED:
            for(k=0; k<t->count; ++k)
            {
               out.write_byte(((byte*)t->value)[k]);
            }
            break;

         case TIFF::type_WORD:
         case TIFF::type_SHORT:
            for(k=0; k<t->count; ++k)
            {
               out.write_word(((word*)t->value)[k]);
            }
            break;

         case TIFF::type_DWORD:
         case TIFF::type_LONG:
         case TIFF::type_FLOAT:
            for(k=0; k<t->count; ++k)
            {
               out.write_dword(((dword*)t->value)[k]);
            }
            break;

         case TIFF::type_RATIONAL:
         case TIFF::type_SRATIONAL:
            for(k=0; k<t->count; ++k)
            {
               out.write_dword(((rational*)t->value)[k].n);
               out.write_dword(((rational*)t->value)[k].d);
            }
            break;
         }
      }
   }
   return true;
}

